#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <GL/glut.h>

#include "lib.h"
#include "Obj.h"
#include "Wall.h"
#include "Canon.h"
#include "Piece.h"
#include "Block.h"

struct _Block {
  int status;
  int track;
  int sector;
  Wall wall;
  ObjList canons;
  ObjList pieces;
};

Block Block_Create(int track, int sector)
{
  Block block;

  block = malloc(sizeof(*block));
  memset(block, 0, sizeof(*block));

  block->status = BLOCK_STATUS_UNCLEAR;
  block->track  = track;
  block->sector = sector;
  block->wall = Wall_Create(block);
  block->canons = ObjList_Create();
  block->pieces = ObjList_Create();

  return block;
}

Block Block_Destroy(Block block)
{
  if (block) {
    if (block->wall) Wall_Destroy(block->wall);
    if (block->canons) ObjList_Destroy(block->canons);
    if (block->pieces) ObjList_Destroy(block->pieces);
    free(block);
  }
  return NULL;
}

int make_pieces(Block block)
{
  Piece piece;
  int i;
  for (i = 0; i < 30; i++) {
    if (block->status == BLOCK_STATUS_CLEARING)
      piece = Piece_Create(block, PIECE_TYPE_SUCCEED);
    else 
      piece = Piece_Create(block, PIECE_TYPE_FAIL);
    Block_AddPiece(block, piece);
  }
  return 0;
}

int Block_Proc(Block block)
{
  Canon canon;
  Piece piece;
  ObjListData obj;

  Wall_Proc(block->wall);

  for (obj = ObjList_GetStart(block->canons);
       !ObjList_IsEndEdge(block->canons, obj);
       obj = ObjListData_GetNext(obj)) {
    canon = (Canon)ObjListData_GetObj(obj);
    if (!Canon_Proc(canon)) {
      obj = ObjList_DeleteObjToPrev(block->canons, obj);
      Block_SetStatus(block, BLOCK_STATUS_CLEARING);
      make_pieces(block);
    }
  }

  for (obj = ObjList_GetStart(block->pieces);
       !ObjList_IsEndEdge(block->pieces, obj);
       obj = ObjListData_GetNext(obj)) {
    piece = (Piece)ObjListData_GetObj(obj);
    if (!Piece_Proc(piece)) {
      obj = ObjList_DeleteObjToPrev(block->pieces, obj);
    }
  }

  return 1;
}

int Block_MakePolygon(Block block, int track_num, int sector_num)
{
  Canon canon;
  Piece piece;
  ObjListData obj;
  double r;
  double angle;
  vector_t position;

  angle = 360.0 * ((double)block->sector / sector_num);
  r = cos(M_PI / sector_num) * (1.0 / (track_num + 1)) * (block->track + 2);

  vector_set(&position,
	     cos((2.0 * M_PI) * (angle / 360.0)) * r,
	     sin((2.0 * M_PI) * (angle / 360.0)) * r,
	     0.0);

  glRotated(angle, 0.0, 0.0, 1.0);
  glTranslated(r, 0.0, 0.0);

  glPushMatrix();
  Wall_MakePolygon(block->wall, track_num, sector_num, block->track);
  glPopMatrix();

  for (obj = ObjList_GetStart(block->canons);
       !ObjList_IsEndEdge(block->canons, obj);
       obj = ObjListData_GetNext(obj)) {
    canon = (Canon)ObjListData_GetObj(obj);
    glPushMatrix();
    Canon_MakePolygon(canon);
    glPopMatrix();
  }

  for (obj = ObjList_GetStart(block->pieces);
       !ObjList_IsEndEdge(block->pieces, obj);
       obj = ObjListData_GetNext(obj)) {
    piece = (Piece)ObjListData_GetObj(obj);
    glPushMatrix();
    Piece_MakePolygon(piece);
    glPopMatrix();
  }

  return 0;
}

BlockStatus_t Block_GetStatus(Block block)
{
  return block->status;
}

BlockStatus_t Block_SetStatus(Block block, BlockStatus_t status)
{
  int valid = 0;

  switch (block->status) {
  case BLOCK_STATUS_UNCLEAR:
    if ((status == BLOCK_STATUS_TRY) ||
	(status == BLOCK_STATUS_ATTACK))
      valid++;
    break;
  case BLOCK_STATUS_TRY:
    if (status == BLOCK_STATUS_CLEARING) {
      if (ObjList_GetLength(block->canons) == 0) {
	status = BLOCK_STATUS_UNCLEAR;
	valid++;
      }
    } else if (status == BLOCK_STATUS_ATTACK) {
      valid++;
    }
    break;
  case BLOCK_STATUS_ATTACK:
    if (status == BLOCK_STATUS_CLEARING) valid++;
    break;
  case BLOCK_STATUS_CLEARING:
    if (status == BLOCK_STATUS_CLEARED) valid++;
    break;
  case BLOCK_STATUS_CLEARED:
  default:
    break;
  }

  if (valid) {
    block->status = status;
  }

  return block->status;
}

int Block_AddCanon(Block block, Canon canon, BlockStatus_t status)
{
  ObjList_InsertObjToEnd(block->canons, canon, (ObjDestructor)Canon_Destroy);
  Block_SetStatus(block, status);
  return 0;
}

int Block_AddPiece(Block block, Piece piece)
{
  ObjList_InsertObjToEnd(block->pieces, piece, (ObjDestructor)Piece_Destroy);
  return 0;
}
